import json
import os
from abc import ABC, abstractmethod
from typing import Dict, List, TypedDict

import pandas as pd

from sziszapangma.model.model import UUIDable


class RelationItem(TypedDict):
    first_id: str
    first_type: str
    second_id: str
    second_type: str


class RelationManager(ABC):
    @abstractmethod
    def get_all_relations_for_item(self, item_id: str) -> List[RelationItem]:
        pass

    @abstractmethod
    def get_item_by_id(self, item_id: str) -> UUIDable:
        pass

    @abstractmethod
    def get_items_by_query(self, query: str) -> List[UUIDable]:
        pass

    @abstractmethod
    def save_item(self, item: UUIDable):
        pass

    @abstractmethod
    def save_relation(self, item1: UUIDable, item2: UUIDable):
        pass

    @abstractmethod
    def commit(self):
        pass

    @abstractmethod
    def get_all_items(self) -> List[UUIDable]:
        pass


class FileRelationManager(RelationManager):
    relations_csv_path: str
    items_json_path: str
    relations_dataframe: pd.DataFrame
    items_dict: Dict[str, UUIDable]

    def __init__(self, relations_csv_path: str, items_json_path: str):
        self.relations_csv_path = relations_csv_path
        self.items_json_path = items_json_path
        if os.path.isfile(relations_csv_path):
            self.relations_dataframe = pd.read_csv(relations_csv_path)
        else:
            self.relations_dataframe = pd.DataFrame(
                [], columns=["first_id", "first_type", "second_id", "second_type"]
            )
        if os.path.isfile(items_json_path):
            with open(items_json_path, "r") as f:
                self.items_dict = json.loads(f.read())
        else:
            self.items_dict = dict()

    def get_all_items(self) -> List[UUIDable]:
        return list(self.items_dict.values())

    def get_items_by_query(self, query: str) -> List[UUIDable]:
        return self.relations_dataframe.query(query).to_dict("records")

    def get_all_relations_for_item(self, item_id: str) -> List[RelationItem]:
        return self.relations_dataframe[self.relations_dataframe["first_id"] == item_id].to_dict(
            "records"
        )

    def get_item_by_id(self, item_id) -> UUIDable:
        return self.items_dict[item_id]

    def save_item(self, item: UUIDable):
        item_id = item.get("id")
        if item_id is not None:
            self.items_dict[item_id] = item

    def save_relation(self, item1: UUIDable, item2: UUIDable):
        relation_1_2 = {
            "first_id": item1["id"],
            "first_type": item1["type"],
            "second_id": item2["id"],
            "second_type": item2["type"],
        }
        relation_2_1 = {
            "first_id": item2["id"],
            "first_type": item2["type"],
            "second_id": item1["id"],
            "second_type": item1["type"],
        }
        self.relations_dataframe = self.relations_dataframe.append(relation_1_2, ignore_index=True)
        self.relations_dataframe = self.relations_dataframe.append(relation_2_1, ignore_index=True)

    def commit(self):
        self.relations_dataframe.to_csv(self.relations_csv_path, index=False)
        items_json = json.dumps(self.items_dict)
        with open(self.items_json_path, "w") as f:
            f.write(items_json)
